/*! \file str.c
 *  \brief Definitions of the string methods.
 *  \author Peter C. Chapin <peter@pchapin.org>
 *  \date November 27, 2017
 */

#include <assert.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "str.h"

// Small enough to be easy on memory. Large enough to be useful.
#define INITIAL_CAPACITY 8

//-----------------------------
//      Internal Functions
//-----------------------------

// The following function returns the smallest power of 2 that is greater than new_size (minimum
// of INITIAL_CAPACITY). This function is used to compute an appropriate new capacity in cases
// where that is necessary. This function makes no provision for overflow.
//
// It is necessary to return a value greater than new_size to allow space for a terminating null
// byte (without requiring any additional allocations) when it is needed.
//
static size_t round_up( size_t new_size )
{
    size_t size = INITIAL_CAPACITY;

    while( size <= new_size ) size <<= 1;
    return size;
}

//-----------------------------
//      External Functions
//-----------------------------


int string_construct( string *object )
{
    object->start    = (char *)malloc( INITIAL_CAPACITY );
    object->size     = 0;
    object->capacity = INITIAL_CAPACITY;
    if( object->start == NULL ) return -1;
    return 0;
}


void string_destroy( string *object )
{
    free( object->start );
    object->start    = NULL;
    object->size     = 0;
    object->capacity = INITIAL_CAPACITY;
}


int string_erase( string *object )
{
    char *temp = (char *)malloc( INITIAL_CAPACITY );
    if( temp == NULL ) return -1;

    free( object->start );
    object->start    = temp;
    object->start[0] = '\0';  // Only needed to keep splint happy.
    object->size     = 0;
    object->capacity = INITIAL_CAPACITY;
    return 0;
}


int string_copy( string *object, const string *other )
{
    size_t  new_capacity;
    char   *temp;

    assert( other->start != NULL );
    assert( object->start != NULL );

    // If I'm big enough to hold the other, just do the copy.
    if ( other->size < object->capacity ) {
        memcpy( object->start, other->start, other->size );
        object->size = other->size;
    }

    // Otherwise, reallocate myself.
    else {
        new_capacity = round_up( other->size + 1 );
        temp = (char *)malloc( new_capacity );
        if( temp == NULL ) return -1;
        free( object->start );
        object->start = temp;
        memcpy( object->start, other->start, other->size );
        object->size = other->size;
        object->capacity = new_capacity;
    }
    return 0;
}


int string_copycharp( string *object, const char *other )
{
    size_t  other_size;
    size_t  new_capacity;
    char   *temp;

    if( other == NULL ) return 0;
    assert( object->start != NULL );

    other_size = strlen( other );

    // If I'm big enough to hold the other, just do the copy.
    if( other_size < object->capacity ) {
        memcpy( object->start, other, other_size );
        object->size = other_size;
    }

    // Otherwise, reallocate myself.
    else {
        new_capacity = round_up( other_size + 1 );
        temp = (char *)malloc( new_capacity );
        if( temp == NULL ) return -1;
        free( object->start );
        object->start = temp;
        memcpy( object->start, other, other_size );
        object->size = other_size;
        object->capacity = new_capacity;
    }
    return 0;
}


int string_copychar( string *object, char other )
{
    assert( object->start != NULL );
    
    // This version makes no attempt to reduce capacity.
    *object->start = other;
    object->size  = 1;
    return 0;
}


int string_copyf( string *object, const char *format, ... )
{
    // This is nasty and hacked. A better version would resize the buffer until it worked and it
    // would then install the buffer directly into the target object (no need to copy it).
    // Something to fix later...

    int     return_value;
    va_list args;
    char   *buffer = (char *)malloc( 1024 );

    if( buffer == NULL ) return -1;

    va_start( args, format );
    return_value = vsnprintf( buffer, 1024, format, args );
    if( return_value >= 1024 ) {
        free( buffer );
        return -1;
    }
    assert( return_value >= 0 );
    return_value = string_copycharp( object, buffer );
    free( buffer );
    va_end( args );

    return return_value;
}


int string_append( string *object, const string *other )
{
    size_t  new_capacity;
    char   *temp;

    assert( other->start != NULL );
    assert( object->start != NULL );

    // If the capacity is sufficient, just append.
    if( object->size + other->size < object->capacity ) {
        memcpy( object->start + object->size, other->start, other->size );
        object->size += other->size;
    }
    
    // Otherwise, reallocate myself.
    else {
        new_capacity = round_up( object->size + other->size + 1 );
        temp = ( char * )malloc( new_capacity );
        if( temp == NULL ) return -1;
        memcpy( temp, object->start, object->size );
        free( object->start );
        object->start = temp;
        memcpy( object->start + object->size, other->start, other->size );
        object->size += other->size;
        object->capacity = new_capacity;
    }
    return 0;
}


int string_appendcharp( string *object, const char *other )
{
    size_t  other_size;
    size_t  new_capacity;
    char   *temp;

    assert( object->start != NULL );
    
    if( other == NULL ) return 0;
    other_size = strlen( other );

    // If the capacity is sufficient, just append.
    if( object->size + other_size < object->capacity ) {
        memcpy( object->start + object->size, other, other_size );
        object->size += other_size;
    }

    // Otherwise, reallocate myself.
    else {
        new_capacity = round_up( object->size + other_size + 1 );
        temp = (char *)malloc( new_capacity );
        if( temp == NULL ) return -1;
        memcpy( temp, object->start, object->size );
        free( object->start );
        object->start = temp;
        memcpy( object->start + object->size, other, other_size );
        object->size += other_size;
        object->capacity = new_capacity;
    }
    return 0;
}


int string_appendchar( string *object, char other )
{
    size_t  new_capacity;
    char   *temp;

    assert( object->start != NULL );

    // If the capacity is sufficient, just append.
    if( object->size + 1 < object->capacity ) {
        *( object->start + object->size ) = other;
        object->size++;
    }

    // Otherwise reallocate myself.
    else {
        new_capacity = round_up( object->size + 1 + 1 );
        temp = (char *)malloc( new_capacity );
        if( temp == NULL ) return -1;
        memcpy( temp, object->start, object->size );
        free( object->start );
        object->start = temp;
        *( object->start + object->size ) = other;
        object->size++;
        object->capacity = new_capacity;
    }
    return 0;
}


int string_appendf( string *object, const char *format, ... )
{
    // This is nasty and hacked. A better version would resize the buffer until it worked. It
    // would be nice if we could also avoid the final copy of the buffer into the target object.

    int     return_value;
    va_list args;
    char   *buffer = (char *)malloc( 1024 );

    if( buffer == NULL ) return -1;

    va_start( args, format );
    return_value = vsnprintf( buffer, 1024, format, args );
    if( return_value >= 1024 ) {
        free( buffer );
        return -1;
    }
    assert( return_value >= 0 );
    return_value = string_appendcharp( object, buffer );
    free( buffer );
    va_end( args );
    
    return return_value;
}


int string_prepend( string *object, const string *other )
{
    char  *temp;
    size_t new_capacity;

    assert( object->start != NULL );
    assert( other ->start != NULL );
    
    // If there is sufficient space, don't allocate a new buffer.
    if( object->capacity - object->size > other->size ) {
        memmove( object->start + other->size, object->start, object->size );
        memcpy( object->start, other->start, other->size );
        object->size += other->size;
    }
    else {
        new_capacity = round_up( object->size + other->size );
        if( ( temp = ( char * )malloc( new_capacity ) ) == NULL ) {
            return -1;
        }
        memcpy( temp, other->start, other->size );
        memcpy( temp + other->size, object->start, object->size );

        // I was able to create the temp workspace correctly... now gut it.
        free( object->start );
        object->start    = temp;
        object->size    += other->size;
        object->capacity = new_capacity;
    }
    return 0;
}


int string_prependcharp( string *object, const char *other )
{
    char  *temp;
    size_t new_capacity;
    size_t other_size;

    assert( object->start != NULL );
    if( other == NULL ) return 0;
    
    other_size = strlen( other );
   
    // If there is sufficient space, don't allocate a new buffer.
    if( object->capacity - object->size > other_size ) {
        memmove( object->start + other_size, object->start, object->size );
        memcpy( object->start, other, other_size );
        object->size += other_size;
    }
    else {
        new_capacity = round_up( object->size + other_size );
        if( ( temp = ( char * )malloc( new_capacity ) ) == NULL ) {
            return -1;
        }
        memcpy( temp, other, other_size );
        memcpy( temp + other_size, object->start, object->size );

        // I was able to create the temp workspace correctly... now gut it.
        free( object->start );
        object->start    = temp;
        object->size    += other_size;
        object->capacity = new_capacity;
    }
    return 0;
}


int string_prependchar( string *object, char other )
{
    char  *temp;
    size_t new_capacity;

    assert( object->start != NULL );    
   
    // If there is sufficient space, don't allocate a new buffer.
    if( object->capacity - object->size > 1 ) {
        memmove( object->start + 1, object->start, object->size );
        *( object->start ) = other;
        object->size += 1;
    }
    else {
        new_capacity = round_up( object->size + 1 );
        if( ( temp = ( char * )malloc( new_capacity ) ) == NULL ) {
            return -1;
        }
        *temp = other;
        memcpy( temp + 1, object->start, object->size );

        // I was able to create the temp workspace correctly... now gut it.
        free( object->start );
        object->start    = temp;
        object->size    += 1;
        object->capacity = new_capacity;
    }
    return 0;
}


size_t string_length( const string *object )
{
  return object->size;
}


char string_getcharat( const string *object, size_t char_index )
{
    assert( object->start != NULL );
    assert( char_index < object->size );
    return *( object->start + char_index );
}


char *string_getcharp( string *object )
{
    assert( object->start != NULL );
    
    *( object->start + object->size ) = '\0';
    return object->start;
}


void string_putcharat(
  string *object, char other, size_t char_index )
{
    assert( object->start != NULL );
    assert( char_index < object->size );
    *( object->start + char_index ) = other;
}


int string_equal( const string *left, const string *right )
{
    assert( left->start != NULL );
    assert( right->start != NULL );
    if( left->size != right->size ) return 0;
    if( memcmp( left->start, right->start, left->size ) != 0 ) return 0;
    return 1;
}


int string_less( const string *left, const string *right )
{
    char   *leftp = left->start;
    char   *rightp = right->start;
    size_t  index;

    assert( left->start != NULL && right->start != NULL );
  
    for( index = 0; index < left->size; index++ ) {
        if( index >= right->size ) break;
        if( *leftp < *rightp ) return 1;
        if( *leftp > *rightp ) return 0;
        leftp++; rightp++;
    }

    // If the right side is longer, then the left side is a prefix.
    if( index < right->size ) {
        return 1;
    }
    return 0;
}


size_t string_findchar( const string *haystack, char needle )
{
    char   *haystackp = haystack->start;
    size_t  index;

    assert( haystack->start != NULL );

    // Rewrite in terms of a memory search function. (Can't use strchr() because strings might
    // contain embedded null bytes).
    // 
    for( index = 0; index < haystack->size; index++ ) {
        if( *haystackp == needle ) return index;
    }
    return (size_t)-1;
}


size_t string_findstring( const string *haystack, const string *needle )
{
    char   *haystackp = haystack->start;
    size_t  index1;
    size_t  index2;

    assert(haystack->start != NULL);
    assert(needle->start != NULL);

    // An empty needle is an error. (Is this right?)
    if( needle->size == 0 ) return 0;

    for( index1 = 0; index1 < (haystack->size - needle->size + 1); index1++ ) {
        char *needlep = needle->start;

        for( index2 = index1; index2 < index1 + needle->size; index2++ ) {
            if( haystackp[index2] != *needlep ) break;
        }

        // Did the inner loop end naturally? If so, there is a match.
        if( index2 >= index1 + needle->size ) return index1;
    }
    return (size_t)-1;
}


int string_substring(
  const string *source,
        string *target,
        size_t        index,
        size_t        length )
{
    size_t temp_index;

    assert( source->start != NULL );
    assert( target->start != NULL );
    assert( index + length < source->size );

    if( string_erase(target) == -1 ) return -1;
    for( temp_index = index; temp_index < index + length; temp_index++ ) {
        if( temp_index >= source->size ) break;
        if( string_appendchar( target, source->start[temp_index] ) == -1 ) return -1;
    }
    return 0;
}


void string_reverse( string *object )
{
    char *left;
    char *right;
    assert( object->start != NULL );

    left = object->start;
    right = (object->start + object->size) - 1;

    while( left < right ) {
        char temp = *left;
        *left = *right;
        *right = temp;
        ++left;
        --right;
    }
}


void string_swap( string *left, string *right )
{
    char  *temp_pointer;
    size_t temp_size;
    assert( left != NULL && right != NULL );

    temp_pointer = left->start;
    left->start = right->start;
    right->start = temp_pointer;

    temp_size = left->size;
    left->size = right->size;
    right->size = temp_size;

    temp_size = left->capacity;
    left->capacity = right->capacity;
    right->capacity = temp_size;
}


int string_read( string *object, FILE *infile )
{
    int ch;
    int ungot_ch;

    assert( object->start != NULL );
    assert( infile != NULL );

    if( string_erase( object ) == -1 ) return -1;

    // Skip leading white space.
    while( ( ch = getc( infile ) ) != EOF ) {
        if( ch != ' ' && ch != '\t' && ch != '\f' && ch != '\n' ) {
            ungot_ch = ungetc( ch, infile );
            assert( ungot_ch != EOF );
            break;
        }
    }

    // Read the next "word"
    while( ( ch = getc( infile ) ) != EOF ) {
        if( ch == ' ' || ch == '\t' || ch == '\f' || ch == '\n' ) {
            ungot_ch = ungetc( ch, infile );
            assert( ungot_ch != EOF );
            break;
        }
        if( string_appendchar(object, (char)ch ) == -1 ) {
            ungot_ch = ungetc( ch, infile );
            assert( ungot_ch != EOF );
            return -1;
        }
    }
    
    return ( ch == EOF && object->size == 0 ) ? -1 : 0;
}


int string_readline( string *object, FILE *infile )
{
    int ch;
    int ungot_ch;

    assert( object->start != NULL );
    assert( infile != NULL );

    if( string_erase( object ) == -1 ) return -1;
    while( ( ch = getc( infile ) ) != EOF ) {
        if( ch == '\n' ) break;
        if( string_appendchar( object, (char)ch ) == -1 ) {
            ungot_ch = ungetc( ch, infile );
            assert( ungot_ch != EOF );
            return -1;
        }
    }

    return ( ch == EOF && object->size == 0 ) ? -1 : 0;
}


int string_write( const string *object, FILE *outfile )
{
    char   *objectp = object->start;
    size_t  index;

    assert( object->start != NULL );
    assert( outfile != NULL );

    for( index = 0; index < object->size; index++ ) {
        if( fputc( *objectp, outfile ) == EOF ) return -1;
        objectp++;
    }

    return 0;
}


int string_writeline( const string *object, FILE *outfile )
{
    assert( object->start != NULL );
    assert( outfile != NULL );

    if( string_write( object, outfile ) == -1 ) return -1;
    if( fputc( '\n', outfile ) == EOF ) return -1;
    return 0;
}
