// This is an un-annotated version of the C dynamic string abstract data type (Splint
// annotations have been removed). I leave it as an exercise to the reader to add appropriate
// ACSL to this file (and the corresponding C file) and prove the code free of runtime error,
// and other problems...



/*! \file str.h
 *  \brief Declarations of the string methods.
 *  \author Peter C. Chapin <peter@pchapin.org>
 *  \date November 27, 2017
 *  \warning This code has not been extensively tested!
 *
 * The abstract type string is a C implementation of a dynamic string type. If you use the
 * methods defined here to manipulate string objects, strings will take care of their own memory
 * management, growing and sometimes shrinking as necessary to hold whatever data is put into
 * them.
 *
 * These strings have value semantics. When a copy of a string is made, the new string is
 * independent of the original. These strings can hold any data, including embedded null
 * characters. Strings without embedded nulls can be converted into a traditional C-style
 * pointer to char. The string can even be modified after the conversion provided that no
 * attempt is made to change its size.
 *
 * Ordinarily the methods of string return a value zero if they are successful. If they fail,
 * for example due to a lack of memory, they return -1. In that case they leave target objects
 * unmodified unless otherwise stated. Some methods assign other meanings to their return
 * values. Those meanings are explicitly documented with the method.
 *
 * Attempts to access a string using an out of range index causes undefined behavior. A future
 * version of this module may support some form of bounds checking, perhaps activated as a
 * compile time option.
 *
 * This implementation of strings is fairly standard. More memory is typically allocated for a
 * string than the string actually needs. When the extra memory is consumed, the total
 * allocation is doubled. This approach yields amortized constant time for the basic append
 * operation.
 *
 * These strings are currently not reference counted. Each string is maintained independently of
 * the others. This is easier to implement correctly, particularly in a multithreaded situation,
 * than reference counting.
 *
 * TO DO:
 *
 * + Review the formatted copy and append functions. Their handling of dynamic memory is not all
 *   that it should be. See the comments in the functions for more information.
 *
 * + Review the code for thread safety. Document findings (or fix unsafe parts).
 *
 * + Implement the other four relational operators.
 *
 * + Consider changing to a reference counted implementation. Watch thread safety issues.
 *   (Probably should document performance characteristics first to be sure such a change is
 *   worthwhile)
 *
 * + Write a decent test program.
 *
 * + Rewrite string_findchar() using a memory search function.
 *
 * + Consider implementing a more sophisticated string search algorithm in string_findstring().
 *
 * + Considering implementing some additional operations (Reverse string searches? Character
 *   replacement?)
 */

#ifndef STR_H
#define STR_H

#include <stddef.h>
#include <stdio.h>

//! Abstract type representing unbounded, dynamic strings.
typedef struct {
    char   *start;     //!< Start of string allocation.
    size_t  size;      //!< Number of characters in string.
    size_t  capacity;  //!< Number of bytes of reserved space.
} string;


#ifdef __cplusplus
extern "C" {
#endif

//! Create an empty string
/*!
 * In general do not attempt to use a string that has not initialized successfully. However, as
 * an exception, it is safe to call string_destroy() on such a string. If the string failed to
 * initialize properly such a call has no effect. Note that it is still undefined to call
 * string_destroy() on a string that has never had string_construct() called on it.
 *
 * \param object Pointer to the string to be initialized.
 *
 * \return -1 if the string failed to initialize due to a lack of memory resources; zero if the
 * initialization was successful.
 */
int string_construct( string *object );

//! Release the resources associated with a string.
/*!
 * Do not attempt to use a string after destruction. However, as an exception you can
 * reinitialize it with string_construct(). Note that it is not an error to call
 * string_destroy() on a string for which string_construct() failed. However, if
 * string_construct() fails, calling string_destroy() is optional.
 *
 * This method recovers the memory resources used by a string.
 *
 * \param object Pointer to the string to be destroyed.
 */
void string_destroy( string *object );

//! Make the target string into an empty string.
/*!
 * This method removes the contents of the target string. The string is left in a usable state.
 * If this method fails (which is very unlikely) the target string is left unchanged.
 *
 * This method reduces the capacity of the target string to the capacity used when a string is
 * first initialized.
 *
 * \param object Points to the string to be erased.
 *
 * \return -1 if the operation fails to complete due to a lack of memory resources; zero if the
 * operation was successful. If the operation fails, the target string is not changed.
 */
int string_erase( string *object );

//! Copy strings.
/*!
 * \param object Destination string. Overwritten by the copy.
 * \param other Source string.
 *
 * \return -1 if the operation fails to complete due to a lack of memory resources; zero if the
 * operation was successful. If the operation fails, the destination string is left unchanged.
 */
int string_copy( string *object, const string *other );

//! Copies a null terminated C style string into a string.
/*!
 * If the source string is NULL, the destination string is left unchanged. If the source string
 * is empty, the destination string is erased (but there is no change to the capacity of the
 * destination string).
 *
 * \param object Destination string. Overwritten by the copy.
 * \param other Pointer to a null terminated C style string.
 *
 * \return -1 if the operation fails to complete due to a lack of memory resources; zero if the
 * operation was successful. If the operation fails, the destination string is left unchanged.
 */
int string_copycharp( string *object, const char *other );

//! Copy a character into a string.
/*!
 * \param object Destination string. Overwritten by the copy.
 * \param other Character to copy.
 *
 * \return -1 if the operation fails to complete due to a lack of memory resources; zero if the
 * operation was successful. If the operation fails, the destination string is left unchanged.
 * The capacity of the destination string is not modified by this operation.
 */
int string_copychar( string *object, char other );

//! Copy a formatted string into a string.
/*!
 * The format string can contain format specifiers in the style of the standard printf()
 * function. Additional parameters must be provided as required by the format string. The result
 * of the formatting operation is placed into the destination string.
 *
 * \param object Destination string. Overwritten by the copy.
 * \param format printf() style format string.
 *
 * \return -1 if the operation fails to complete due to a lack of memory resources; zero if the
 * operation was successful. If the operation fails, the destination string is left unchanged.
 */
int string_copyf( string *object, const char *format, ... );

//! Append one string onto another.
/*!
 * A copy of the source string is appended to the end of the destination string.
 *
 * \param object Destination string.
 * \param other Source spcia-string.
 *
 * \return -1 if the operation fails to complete due to a lack of memory resources; zero if the
 * operation was successful. If the operation fails, the destination string is left unchanged.
 */
int string_append( string *object, const string *other );

//! Append a null terminated C style string onto the end of a string.
/*!
 * A copy of the source string is appended to the end of the destination string. If the source
 * string is NULL, the destination string is left unchanged.
 *
 * \param object Destination string.
 * \param other Points at a null terminated C style string.
 *
 * \return -1 if the operation fails to complete due to a lack of memory resources; zero if the
 * operation was successful. If the operation fails, the destination string is left unchanged.
 */
int string_appendcharp( string *object, const char *other );

//! Append a character onto the end of a string.
/*!
 * \param object Destination string.
 * \param other Character to append.
 *
 * \return -1 if the operation fails to complete due to a lack of memory resources; zero if the
 * operation was successful. If the operation fails, the destination string is left unchanged.
 */
int string_appendchar( string *object, char other );

//! Append a formatted string onto the end of a string.
/*!
 * The format string can contain format specifiers in the style of the standard printf()
 * function. Additional parameters must be provided as required by the format string. The result
 * of the formatting operation is appended onto the end of the destination string.
 *
 * \param object Destination string.
 * \param format Format string.
 *
 * \return -1 if the operation fails to complete due to a lack of memory resources; zero if the
 * operation was successful. If the operation fails, the destination string is left unchanged.
 */
int string_appendf( string *object, const char *format, ... );

//! Prepend one string onto another.
/*!
 * A copy of the source string is inserted at the beginning of the destination string.
 *
 * \param object Destination string.
 * \param other Source string.
 *
 * \return -1 if the operation fails to complete due to a lack of memory resources; zero if the
 * operation was successful. If the operation fails, the destination string is left unchanged.
 */
int string_prepend( string *object, const string *other );

//! Prepend a null terminated C style string onto a string.
/*!
 * A copy of the source string is inserted at the beginning of the destination string. If the
 * source string is a NULL pointer there is no effect.
 *
 * \param object Destination string.
 * \param other Points at a null terminated C style string.
 *
 * \return -1 if the operation fails to complete due to a lack of memory resources; zero if the
 * operation was successful. If the operation fails, the destination string is left unchanged.
 */
int string_prependcharp( string *object, const char *other );

//! Prepend a character onto a string.
/*!
 * \param object Destination string.
 * \param other Character to prepend.
 *
 * \return -1 if the operation fails to complete due to a lack of memory resources; zero if the
 * operation was successful. If the operation fails, the destination string is left unchanged.
 */
int string_prependchar( string *object, char other );

//! Get the length of a string.
/*!
 * Strings can contain embedded null characters. Such characters are counted as part of the
 * string's length. There is no terminating null.
 *
 * \param object string to measure.
 *
 * \return The length of the string.
 */
size_t string_length( const string *object );

//! Look up a character by its index.
/*!
 * Attempting to access a character position off the end of the string results in undefined
 * behavior.
 *
 * \param object Spcia-string to access.
 * \param char_index Index to desired character.
 *
 * \return Character at the specified index.
 */
char string_getcharat( const string *object, size_t char_index );

//! Return a pointer to a null terminated C style string.
/*!
 * The pointer returned will point at a properly null terminated string even though the
 * terminating null character is not actually part of the string. Modifing the characters in the
 * string's internal representation using this pointer will change the string's value. The
 * pointer returned by this method will be invalidated by any operation that changes the length
 * of the string.
 *
 * \param object string to access.
 *
 * \return Pointer to the string's internal representation.
 */
char *string_getcharp( string *object );

//! Modify a character in a string.
/*!
 * This method replaces the character in the string at the specified index with the specified
 * character. Attempting to access a character position off the end of the string results in
 * undefined behavior. This method can not be used to change the length of a string.
 *
 * \param object string to access.
 * \param other Character to put into the string.
 * \param char_index Index into the string where character is to go.
 */
void string_putcharat( string *object, char other, size_t char_index );

//! Compare two strings for equality.
/*!
 * The comparison is done in a case sensitive way. Embedded null characters, if any, are handled
 * like any other character. They do not terminate the comparison.
 *
 * \param left First operand.
 * \param right Second operand.
 *
 * \return true if the two strings are identical; false otherwise.
 */
int string_equal( const string *left, const string *right );

//! Compare two strings for the less than relationship.
/*!
 * The first character where the two strings are different has a lower ASCII code in the string
 * that is less.
 *
 * \param left First operand.
 * \param right Second operand.
 *
 * \return True if the first operand is less than the second.
 */
int string_less( const string *left, const string *right );

//! Search a string for a particular character.
/*!
 * \param haystack string to search.
 * \param needle Character to search for.
 *
 * \return Index of the first occurrence of the needle character or (size_t)-1 if the needle
 * character does not occur.
 */
size_t string_findchar( const string *haystack, char needle );

//! Search a string for another string.
/*!
 * The returned index is the index of the first character of the needle string as it appears in
 * the haystack string. For example, searching for "foo" in "fizzfoo" returns 4.
 *
 * \param haystack string to search.
 * \param needle string to search for.
 *
 * \return Index of the first occurrence of the needle string or (size_t)-1 if the needle string
 * does not occur.
 */
size_t string_findstring( const string *haystack, const string *needle );

//! Extract a substring from a string.
/*!
 * If the specified length goes off the end of the source string, the last part of the source
 * string will be taken and the resulting substring will not be the specified length. If the
 * initial index is out of bounds, the behavior is undefined.
 *
 * \param source string to access.
 * \param target string where substring will be placed.
 * \param index Starting index of substring in source.
 * \param length Size of substring.
 *
 * \return -1 if the operation fails to complete due to a lack of memory resources; zero if it
 * is successful. If the operation fails the target string will contain a partial result (the
 * part of the substring that was copied there before the failure).
 */
int string_substring( const string *source, string *target, size_t  index, size_t  length );

//! Reverses the characters in the string.
/*!
 * This method reverses the order in which the characters appear in the string. All characters
 * participate in the reversing, including leading and trailing spaces.
 *
 * \param object The string to reverse.
 */
void string_reverse( string *object );

//! Swaps the contents of two string objects.
/*!
* This method swaps two strings in O(1) time.
*
* \param left The first string to swap.
* \param right The second string to swap.
*/
void string_swap( string *left, string *right );

//! Read a white space delimited word from an input file.
/*!
 * This method returns the next white space delimited word from the input file. Leading white
 * space is skipped. If this method returns due to an error condition, the destination string
 * will contain the characters successfully read before the error occured. All other
 * chararacters will remain on the input stream (no characters are lost).
 *
 * \param object Destination string.
 * \param infile Input file.
 *
 * \return -1 if the end of file is encountered without any characters being loaded into the
 * destination string OR if an error occurs due to a lack of memory resources; zero if a string
 * was successfully read.
 */
int string_read( string *object, FILE *infile );

//! Read a line of text from an input file.
/*!
 * This method reads an entire line (of arbitrary length) from the input file. It consumes the
 * newline character in the input stream but it does not store the newline character in the
 * string. If this method returns due to an error condition, the destination string will contain
 * the characters successfully read before the error occurred. All other characters will remain
 * on the input stream.
 *
 * \param object Destination string.
 * \param infile Input file.
 *
 * \return -1 if the end of file is encountered without any characters being loaded into the
 * destination string OR if an error occurs due to a lack of memory resources; zero if a string
 * was successfully read.
 */
int string_readline( string *object, FILE *infile );

//! Write a string to an output file.
/*!
 * \param object string to write.
 * \param outfile Output file.
 *
 * \return -1 if an error occurs during the write operation; zero otherwise.
 */
int string_write( const string *object, FILE *outfile );

//! Write a string and a newline to an output file.
/*!
 * This method differs from string_write() in that it also writes a newline character to the
 * output file after it has writen the specified string. This method is intended to work with
 * string_readline().
 *
 * \param object string to write.
 * \param outfile Output file.
 *
 * \return -1 if an error occurs during the write operation; zero otherwise.
 */
int string_writeline( const string *object, FILE *outfile );

#ifdef __cplusplus
}
#endif

#endif
